@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-import sys
+import sys, os
 import pickle, time
 from copy import deepcopy
 from shutil import copy
@@ -25,16 +25,43 @@ import tqdm
 
 
 # Trace Normalization
-def matrix_entropy(W):
-    """Matrix entropy of W real rectangular matrix, computed using the singular values; may be slow"""
-    W = W / np.trace(W)
-    m = W.shape[1]
-    u, sv, vh = svd(W)
-    
-    rank = np.linalg.matrix_rank(W)
-    p = sv * sv
-    p = p / np.sum(p)
-    
+#def matrix_entropy_(W):
+#   """Matrix entropy of W real rectangular matrix, computed using the singular values; may be slow"""
+#    W = W / np.trace(W)
+#    N, M  = np.min(W.shape), np.max(W.shape)
+#    u, sv, vh = svd(W)  
+#    
+#    evals  = sv * sv
+#    return matrix_entropy(evals, N)
+
+
+def matrix_rank(svals, N, tol=None):
+    """Matrix rank, computed from the singular values directly
+
+    svals are the singular values
+    N is the largest dimension of the matrix
+ 
+    re-implements np.linalg.matrix_rank(W) """
+    S = svals
+    if tol is None:
+        tol = np.max(S) * N * np.finfo(S.dtype).eps
+    return np.count_nonzero(S > tol)
+
+
+def calc_rank_loss(svals, N, tol=None):
+    """Rank loss for this matrix, from the singular values and the largest dim N."""
+
+    rank = matrix_rank(svals, N, tol=tol) #np.linalg.matrix_rank(W)
+    return len(svals) - rank
+
+def matrix_entropy(svals, N):
+    """Matrix entropy of real, computed using the singular values, and the dim N"""
+
+
+    rank = matrix_rank(svals, N) #np.linalg.matrix_rank(W)
+
+    evals = svals*svals
+    p = evals / np.sum(evals)
     if (rank == 1):
         rank = 1.000001
     entropy = -np.sum(p * np.log(p)) / np.log(rank) 
@@ -103,7 +130,6 @@ def hard_rank(W, sv):
 def get_shuffled_eigenvalues(W, layer=7, num=100):
     "get eigenvalues for this model, but shuffled, num times"
     
-    print("get_shuffled_eigenvalues")
     N, M = W.shape[0], W.shape[1]       
 
     if (N < M):
@@ -123,9 +149,9 @@ def get_shuffled_eigenvalues(W, layer=7, num=100):
     return evals
 
 
-def plot_density_and_fit(eigenvalues=None, model=None, layer="",
+def plot_density_and_fit(eigenvalues=None, model=None, layer_name="", layer_id=0,
                      Q=1.0, num_spikes=0, sigma=None,
-                     alpha=0.25, color='blue', skip=False, verbose=True, plot=True):
+                     alpha=0.25, color='blue', skip=False, verbose=True, plot=True, cutoff=0.0):
     """Plot histogram of eigenvalues, for Q, and fit Marchenk Pastur.  
     If no sigma, calculates from maximum eigenvalue (minus spikes)
     Can read keras weights from model if specified.  Does not read PyTorch
@@ -137,6 +163,7 @@ def plot_density_and_fit(eigenvalues=None, model=None, layer="",
    
     if Q == 1:
         to_fit = np.sqrt(eigenvalues)
+        cutoff = np.sqrt(cutoff)
         label = r"$\rho_{emp}(\nu)$"
         title = " W{} SSD, QC Sigma={:0.3}" 
     else:
@@ -147,6 +174,9 @@ def plot_density_and_fit(eigenvalues=None, model=None, layer="",
     if plot:
         plt.hist(to_fit, bins=100, alpha=alpha, color=color, density=True, label=label);
         plt.legend()
+        
+        if cutoff > 0.0:
+            plt.axvline(x=cutoff, linewidth=1, color='r', ls='dashed')
     
     if skip:
         return
@@ -170,7 +200,7 @@ def plot_density_and_fit(eigenvalues=None, model=None, layer="",
         x, mp = marchenko_pastur_pdf(x_min, x_max, Q, sigma)
 
     if plot:
-        plt.title(title.format(layer, sigma))
+        plt.title(title.format(layer_name, sigma))
         plt.plot(x, mp, linewidth=1, color='r', label="MP fit")
         
     if verbose:
@@ -180,20 +210,27 @@ def plot_density_and_fit(eigenvalues=None, model=None, layer="",
     return sigma
 
 
-def plot_density(to_plot, sigma, Q, method="MP"):
+def plot_density(to_plot, sigma, Q, method="MP", color='blue', cutoff=0.0):
     """Method = 'MP' or 'QC'
     
     """
-    to_plot = np.sort(to_plot)
-    x_min, x_max = 0, np.max(to_plot)
-    
+
     if method == "MP":
+        to_plot = np.sort(to_plot)
+        x_min, x_max = 0, np.max(to_plot)
         x, y = marchenko_pastur_pdf(x_min, x_max, Q, sigma)
     elif method == "QC":
+        cutoff = np.sqrt(cutoff)
+        to_plot = np.sort(to_plot)
+        x_min, x_max = 0, np.max(to_plot)
         x, y = quarter_circle_pdf(x_min, x_max, sigma)
         
-    plt.hist(to_plot, bins=100, alpha=0.6, color='blue', density=True)
-    plt.plot(x, y, linewidth=1, color='r')  # , label = method + " fit")
+    plt.hist(to_plot, bins=100, alpha=0.6, color=color, density=True, label="ead")
+    plt.plot(x, y, linewidth=1, color='r', label = method + " fit")
+    plt.legend()
+    
+    if cutoff > 0.0:
+        plt.axvline(x=cutoff, linewidth=1, color='r', ls='dashed')
     
     return None
 
@@ -286,6 +323,8 @@ def vector_entropy(u):
 
 def discrete_entropy(vec, num_bins=100):
     """compute the discrete vector entropy  for this numpy vector, given the number of bins"""
+
+    vec = np.absolute(vec)
     vec = vec - np.mean(vec)
     h = np.histogram(vec, density=True, bins=num_bins)[0];
     p = np.array(h) + 0.0000000001
@@ -340,25 +379,28 @@ def best_dist(fit):
     distName = 'power_law'
     dist = "PL"
 
-    R, p = fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
-    if R > 0 and p <= 0.05:
-        distName = 'truncated_power_law'
-        dist = 'TPL'
-        
-    R, p = fit.distribution_compare(distName, 'exponential', normalized_ratio=True)
-    if R < 0 and p <= 0.05:
-        dist = 'EXP'
-        return dist
-
-    R, p = fit.distribution_compare(distName, 'stretched_exponential', normalized_ratio=True)
-    if R < 0 and p <= 0.05:
-        dist = 'S_EXP'
-        return dist
-        
-    R, p = fit.distribution_compare(distName, 'lognormal', normalized_ratio=True)
-    if R < 0 and p <= 0.05:
-        dist = 'LOG_N'
-        return dist
+    with warnings.catch_warnings():
+        warnings.simplefilter(action='ignore', category=RuntimeWarning)
+    
+        R, p = fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
+        if R > 0 and p <= 0.05:
+            distName = 'truncated_power_law'
+            dist = 'TPL'
+            
+        R, p = fit.distribution_compare(distName, 'exponential', normalized_ratio=True)
+        if R < 0 and p <= 0.05:
+            dist = 'EXP'
+            return dist
+    
+        R, p = fit.distribution_compare(distName, 'stretched_exponential', normalized_ratio=True)
+        if R < 0 and p <= 0.05:
+            dist = 'S_EXP'
+            return dist
+            
+        R, p = fit.distribution_compare(distName, 'lognormal', normalized_ratio=True)
+        if R < 0 and p <= 0.05:
+            dist = 'LOG_N'
+            return dist
 
     return dist
 
@@ -413,7 +455,15 @@ def resid_mp(p, evals, Q, bw, allresid=True, num_spikes=0, debug=False):
     X_plot = xde[:, np.newaxis]
     log_dens = kde.score_samples(X_plot)
     yde = np.exp(log_dens)   
-    
+
+    # hack to try to fix fits
+    # remove areas where the density is almost zero
+    THRESH = 0.01
+    bad_ids = np.where(np.array(yde) < THRESH)[0]
+    good_ids = np.where(np.array(yde) > THRESH)[0]
+    xde = np.array(xde)[good_ids]
+    yde = np.array(yde)[good_ids]
+
     if Q == 1:
         # Quarter Cirle Law fit for this sigma
         xmp, ymp = quarter_circle_fun(xde, sigma=sigma)
@@ -422,20 +472,26 @@ def resid_mp(p, evals, Q, bw, allresid=True, num_spikes=0, debug=False):
         # MP fit for this sigma
         xmp, ymp, a, b = marchenko_pastur_fun(xde, Q=Q, sigma=sigma)
 
-        if (b > max(xde)) | (a > xde[np.where(yde == max(yde))][0]) | (sigma > 1):
-            # print(xde[np.where(yde == max(yde))])
+        ### issue #60  change if statement
+        #if (b > max(xde)) | (a > xde[np.where(yde == max(yde))][0]) | (sigma > 1):
+        if (b < xde[np.where(yde == max(yde))][0] ) | (b > max(xde)) | (a > xde[np.where(yde == max(yde))][0]):
             resid = np.zeros(len(yde)) + 1000
         else:
             # form residual, remove nan's 
             resid = ymp - yde
     #     resid = np.nan_to_num(resid)
-    
+
     if debug:
-        plt.plot(xde, yde)
-        plt.plot(xmp, ymp)
-        plt.show()
+        plt.plot(xde, yde, color='cyan')
+        plt.plot(xmp, ymp, color='orange')
+        plt.axhline(y=THRESH)
+        plt.show(); plt.clf()
         print("sigma {}  mean residual {}".format(sigma, np.mean(resid)))
 
+    # hack to try to fix fits
+    bad_zeros = np.zeros(len(bad_ids))
+    resid = np.concatenate((resid,bad_zeros))
+    
     if allresid:
         return resid
     else:
@@ -473,11 +529,17 @@ def fit_density(evals, Q, bw=0.1, sigma0=None):
     return sigma1, infodict['fvec']
 
 
-def fit_density_with_range(evals, Q, bw=0.1, sigma_range=(slice(0.3, 1.05, 0.1),)):
-    
+def fit_density_with_range(evals, Q, bw=0.1, sigma_range=(slice(0.1, 1.25, 0.01),) ):
     assert type(sigma_range) == tuple, ValueError("sigma_range must be tuple")
     assert type(sigma_range[0]) == slice
-    
+
+    ### issue #60 fix
+    #   reset sigma range
+    sigma_scaling_factor = calc_sigma(Q, evals)
+    sigma_range = (slice(0.05 * sigma_scaling_factor, 2.0 * sigma_scaling_factor, 0.01 * sigma_scaling_factor),)
+    #
+    ###
+
     if Q == 1:
         to_fit = np.sqrt(evals)
     else:
@@ -513,3 +575,173 @@ def plot_loghist(x, bins, xmin):
 
     plt.xscale('log')
     
+    
+def permute_matrix(W):
+    """permute a matrix in a reversible way"""
+    
+    num_params = np.prod(W.shape)
+    vec = W.reshape(num_params)
+    p_ids = np.random.permutation(np.arange(num_params))
+    p_vec = vec[p_ids]
+    p_W = p_vec.reshape(W.shape)
+            
+    return p_W, p_ids
+
+
+def unpermute_matrix(W, p_ids):
+    """unpermute a matrix, using the original ids to permute it"""
+    
+    num_params = np.prod(W.shape)
+    vec = W.reshape(num_params)
+    unp_ids = np.argsort(p_ids)
+    unp_vec = vec[unp_ids]
+    unp_W = unp_vec.reshape(W.shape)
+    
+    return unp_W
+
+
+def save_fig(plt, figname, layer_id, savedir):
+    """Save the figure to the savedir directory. 
+       If directory is not present, create it
+       """
+    
+    figname = "{}/ww.layer{}.{}.png".format(savedir, layer_id, figname)
+    if not os.path.isdir(savedir):
+        os.mkdir(savedir)
+    plt.tight_layout()
+    plt.savefig(figname)
+    return 
+
+#TODO:
+# check alpha is decreasing
+def fit_clipped_powerlaw(evals, xmin=None, verbose=False, min_alpha=5.0):
+    """Fits a powerlaw only, not a truncated power law
+       clips off the max evals until a powerlaw is found, or stops half-way into the ESD
+       
+       Used to remove power-law tail fingers, which may result from finite size effects
+       
+       Assumes eval are in sort order
+       
+       Does not allow alpha to increase; only activates if alpha < min_alpha """
+       
+    fit = None
+    with warnings.catch_warnings():
+        warnings.simplefilter(action='ignore', category=RuntimeWarning)
+            
+        assert(evals[-1]> evals[0])  
+        N = int(len(evals)/4)
+        xmax = np.max(evals)
+       
+        if xmin is not None and xmin != -1:
+            prev_fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+        else:
+            prev_fit = powerlaw.Fit(evals, xmax=xmax, verbose=verbose)
+            
+        prev_alpha = prev_fit.alpha
+        first_fit = prev_fit
+        
+        for idx in range(1,N):
+            xmax = np.max(evals[-idx])
+             
+            if xmin is not None and xmin != -1:
+                fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+            else:
+                fit = powerlaw.Fit(evals, xmax=xmax, verbose=verbose)
+                
+            print("fit alpha",fit.alpha)  
+            
+            if fit.alpha > prev_alpha:
+                fit = prev_fit
+                break
+            
+            if fit.alpha <= min_alpha: 
+                print("stopping")  
+                break
+            
+            # stop when distribution becomes power law
+            R, p = fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
+            if R > 0.0:
+                break
+            
+            prev_fit = fit
+            prev_alpha = fit.alpha
+                
+        if idx == N:
+            fit = first_fit
+     
+    return fit
+             
+        
+# https://medium.com/@sourcedexter/how-to-find-the-similarity-between-two-probability-distributions-using-python-a7546e90a08d
+def jensen_shannon_distance(p, q):
+    """
+    method to compute the Jenson-Shannon Distance 
+    between two probability distributions
+    """
+
+    # convert the vectors into numpy arrays in case that they aren't
+    p = np.array(p)
+    q = np.array(q)
+
+    # calculate m
+    m = (p + q) / 2
+
+    # compute Jensen Shannon Divergence
+    divergence = (sp.stats.entropy(p, m) + sp.stats.entropy(q, m)) / 2
+
+    # compute the Jensen Shannon Distance
+    distance = np.sqrt(divergence)
+
+    return distance
+
+
+def rescale_eigenvalues(evals):
+    """Rescale eigenvalues by their Frobenius Norm"""
+    
+    M = len(evals)
+    Wnorm = np.sqrt(np.sum(evals))
+    Wscale = np.sqrt(M)/Wnorm
+    evals = (Wscale*Wscale)*evals
+    
+    return evals   
+
+def detX_constraint(evals, rescale=True):
+    """Identifies the number of eigenvalues necessary to best satisify the det X = 1 constraint
+    
+    Parameters
+    ----------
+    evals : float array
+        Eigenvalues, assumed to be in sort order, evals[0]<evals[-1]
+        
+    rescale: bool
+        Rescale the eigenvalues by the Frobenius Norm if not already rescaled
+    
+    Returns:
+    --------
+    
+        num_evals:  int
+        number of eigenvalues in the tail
+        
+        det_x_idx:  int
+        index of the first eigenvalue in the tail
+    
+    """
+
+    assert(evals[0]<evals[-1])
+    num_evals = len(evals)
+    idx = 0
+    
+
+    if rescale:
+        Wnorm = np.sqrt(np.sum(evals))
+        Wscale = np.sqrt(M)/Wnorm
+        evals = (Wscale*Wscale)*evals
+
+    for idx in range(len(evals)-1, 0, -1):
+        detX = np.prod(evals[idx:])
+        if detX < 1.0:
+            num_evals = len(evals) - idx 
+            break    
+    
+    
+    return num_evals, idx
